from typing import Dict, List
from .base import Check
from .issues import Issues, Issue
import bs4


class DuplicateIDCheck(Check):
    def __init__(self, html_soup: bs4.BeautifulSoup) -> None:
        self.html_soup = html_soup

    def run(self) -> Issues:
        issues = Issues()

        ids_seen: Dict[str, List[bs4.element.Tag]] = {}

        for el in self.html_soup.findAll():
            id = el.get("id")
            if not id:
                continue

            if id in ids_seen:
                ids_seen[id].append(el)
            else:
                ids_seen[id] = [el]

        for id, elements in ids_seen.items():
            if len(elements) > 1:
                issues.append(
                    Issue(
                        type="duplicate-id",
                        description="The same `id` should not be used on a page more than once.",
                        context={"id": id, "elements": [str(x) for x in elements]},
                    )
                )

        return issues



from typing import List, Optional
from jinja2 import meta, Environment


def get_references_in_path(path: str, jinja_env: Environment) -> List[str]:
    """Get all (recursive) references that go into this file"""

    with open(path, "r") as f:
        ast = jinja_env.parse(f.read())

    these_references = list(meta.find_referenced_templates(ast))

    references = set()

    for ref in these_references:
        if ref in references or not ref:
            continue
        else:
            references.add(ref)
            reference_path = get_path_for_reference(ref, jinja_env)
            if reference_path:
                references |= set(
                    get_references_in_path(
                        reference_path,
                        jinja_env,
                    )
                )

    return list(references)


def get_path_for_reference(reference: str, jinja_env: Environment) -> Optional[str]:
    return jinja_env.get_template(reference).filename



import bs4
from urllib.parse import urlparse

from .base import Check
from .issues import Issues, Issue


def url_is_valid_absolute(url: str) -> bool:
    parsed = urlparse(url)

    if parsed.netloc.startswith("127.0.0.1"):
        # TODO ideally this would only be allowed in local development
        # https not required
        return True

    if not parsed.netloc:
        # domain missing, probably broken url like "https:///"
        return False

    if parsed.scheme != "https":
        # must be https
        return False

    return True


class BaseOpenGraphCheck(Check):
    og_property = ""

    def __init__(self, html_soup: bs4.BeautifulSoup) -> None:
        self.html_soup = html_soup

    def run(self) -> Issues:
        issues = Issues()

        meta = self.html_soup.find("meta", {"property": f"og:{self.og_property}"})
        self.meta_tag_content = meta.get("content", "") if meta else ""
        if not self.meta_tag_content:
            property_slug = self.og_property.replace("_", "-")
            issues.append(
                Issue(
                    type=f"open-graph-{property_slug}-missing",
                    description=f"The og:{property_slug} meta tag is missing.",
                )
            )

        return issues


class OpenGraphTitleCheck(BaseOpenGraphCheck):
    og_property = "title"


class OpenGraphDescriptionCheck(BaseOpenGraphCheck):
    og_property = "description"

    def run(self) -> Issues:
        issues = super().run()
        meta = self.html_soup.find("meta", {"name": "description"})
        if meta and meta.get("content", ""):
            # if there is a meta description, that is a fine alternative
            # to the more specific og:description
            issues = Issues()
        return issues


class OpenGraphTypeCheck(BaseOpenGraphCheck):
    og_property = "type"


class OpenGraphURLCheck(BaseOpenGraphCheck):
    og_property = "url"

    def run(self) -> Issues:
        issues = super().run()

        url = self.meta_tag_content

        if url:
            # try:
            #     response = fetch.head(url)
            #     if response.status_code != 405:
            #         # just skip for now if HEAD not allowed
            #         response.raise_for_status()
            # except Exception as e:
            #     issues.append(
            #         Issue(
            #             type="open-graph-url-broken",

            #             context={"exception": str(e)},
            #         )
            #     )

            if not url_is_valid_absolute(url):
                issues.append(
                    Issue(
                        type="open-graph-url-not-canonical-https",
                        description="The og:url should be an absolute, HTTPS url.",
                        context={"content": url},
                    )
                )

        return issues


class OpenGraphImageCheck(BaseOpenGraphCheck):
    og_property = "image"

    def run(self) -> Issues:
        issues = super().run()

        url = self.meta_tag_content

        if url:
            # try:
            #     response = fetch.head(url)
            #     if response.status_code != 405:
            #         # just skip for now if HEAD not allowed
            #         response.raise_for_status()
            # except Exception as e:
            #     issues.append(
            #         Issue(
            #             type="open-graph-image-broken", context={"exception": str(e)},
            #         )
            #     )

            if not url_is_valid_absolute(url):
                issues.append(
                    Issue(
                        type="open-graph-image-not-canonical-https",
                        description="The og:image should be an absolute, HTTPS url.",
                        context={"content": url},
                    )
                )

        return issues


class OpenGraphSiteNameCheck(BaseOpenGraphCheck):
    og_property = "site_name"



import os

from .base import Check
from .issues import Issues, Issue


MAX_FILE_SIZES = {
    ".png": 10_000_000,  # 10 MB
    ".jpg": 10_000_000,  # 10 MB
    ".jpeg": 10_000_000,  # 10 MB
}


def sizeof_fmt(num: float) -> str:
    for unit in [" bytes", "KB", "MB", "GB"]:
        if abs(num) < 1000.0:
            return "%3.1f%s" % (num, unit)
        num /= 1000.0
    return "%.1f%s" % (num, "Yi")


class FileSizeCheck(Check):
    def __init__(self, path: str) -> None:
        self.path = path

    def run(self) -> Issues:
        issues = Issues()

        _, ext = os.path.splitext(self.path)

        if ext in MAX_FILE_SIZES:
            size = os.path.getsize(self.path)
            max_size = MAX_FILE_SIZES[ext]
            if size > max_size:
                issues.append(
                    Issue(
                        type=f"file-size-too-large",
                        description=f"Files of type {ext} shouldn't be bigger than {sizeof_fmt(max_size)}",
                        context={
                            "output_path": os.path.relpath(self.path),
                            "file_size": sizeof_fmt(size),
                        },
                    )
                )

        return issues



import logging
import os
import sys
from typing import List

import click
import pygments
import cls_client
import barrel
from honcho.manager import Manager as HonchoManager
from honcho.printer import Printer as HonchoPrinter
from repaint import Repaint

from .core import Combine
from .logger import logger
from .dev import Watcher, Server
from .exceptions import BuildError
from . import __version__


cls_client.set_project_key("cls_pk_QFp5bJFR1RXauHdvRUDpDngE")
cls_client.set_project_slug("combine")
cls_client.set_version(__version__)
cls_client.set_noninteractive_tracking_enabled(True)


@click.group()
@click.version_option(__version__)
@click.pass_context
def cli(ctx: click.Context) -> None:
    pass


@cli.command()
@click.option("--check", is_flag=True, default=False)
@click.option("--env", default="production")
@click.option("--var", multiple=True, default=[])
@click.option("--debug", is_flag=True, default=False)
@click.pass_context
@cls_client.track_command(
    include_kwargs=["check", "env"],
    include_env=["NETLIFY", "CIRCLECI", "TRAVIS", "GITLAB_CI", "GITHUB_ACTIONS", "CI"],
)
def build(
    ctx: click.Context, check: bool, env: str, var: List[str], debug: bool
) -> None:
    """Build the site (typically during deployment)"""
    if debug:
        logger.setLevel(logging.DEBUG)

    variables = dict(x.split("=") for x in var)
    config_path = os.path.abspath("combine.yml")
    combine = Combine(config_path=config_path, env=env, variables=variables)

    click.secho("Building site", bold=True, color=True)
    try:
        combine.build(check=check)
    except BuildError:
        click.secho("Build error (see above)", fg="red", color=True)
        exit(1)

    if check:
        if combine.issues:
            click.secho(
                f"{len(combine.issues)} check{'s' if len(combine.issues) > 1 else ''} failed",
                fg="red",
                color=True,
            )
            exit(1)
        else:
            click.secho("✓ All checks passed", fg="green", color=True)


@cli.command()
@click.option("--port", type=int, default=8000)
@click.option("--debug", is_flag=True, default=False)
@click.option("--repaint", is_flag=True, default=True)
@click.pass_context
def work(ctx: click.Context, port: int, debug: bool, repaint: bool) -> None:
    """Start a local server to build the site while you work"""
    if debug:
        logger.setLevel(logging.DEBUG)

    cls_client.track_event(slug="work", type="command", metadata={}, dispatch=True)

    config_path = os.path.abspath("combine.yml")
    combine = Combine(
        config_path=config_path,
        env="development",
        variables={"base_url": f"http://127.0.0.1:{port}"},
    )
    click.secho("Building site", bold=True, color=True)
    try:
        combine.build(check=True)
    except BuildError:
        click.secho("Build error (see above)", fg="red", color=True)
        exit(1)

    header = (
        """
┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
┃      ┏━━━━━┓                                               ┃
┃      ┗┓   ┏┻━━━┓    Site is live: http://127.0.0.1:%s    ┃
┃    ┏━━┫   ┣━━┓ ┃                                           ┃
┃    ┃ ┏┻━━━┻┓ ┃      Docs: https://combine.dropseed.dev     ┃
┃ ┏━━┻━┻━━━━━┻━┻━━┓                                          ┃
┃ ┣━━━━━━━━━━━━━━━┫   Watching for file changes...           ┃
┃ ┗━━━━◡◡━━━◡◡━━━━┛                                          ┃
┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛
"""
        % port
    )

    click.secho(header, fg="green", bold=True, color=True)

    debug_flag = "--debug" if debug else ""
    repaint_flag = "--repaint" if repaint else ""

    bin_path = os.path.dirname(sys.executable)
    combine_path = os.path.join(bin_path, "combine")
    honcho_env = {
        **os.environ,
        "PYTHONUNBUFFERED": "1",
    }

    manager = HonchoManager(HonchoPrinter())
    manager._system_print = lambda x: None
    manager.add_process(
        "server",
        f"{combine_path} utils server --port {port} {debug_flag} {repaint_flag}",
        env=honcho_env,
    )
    manager.add_process(
        "combine",
        f"{combine_path} utils watch --port {port} {debug_flag} {repaint_flag}",
        env=honcho_env,
    )

    if repaint:
        manager.add_process(
            "repaint",
            f"{os.path.join(bin_path, 'repaint')} serve {'--quiet' if not debug else ''}",
            env=honcho_env,
        )

    # Add additional custom watch processes
    for i, step in enumerate(combine.config.steps):
        if step.has_watch_process:
            name = step.get_name() or f"step-{i}"
            manager.add_process(
                name,
                step.watch,
            )

    manager.loop()


@cli.group()
@click.pass_context
def utils(ctx: click.Context) -> None:
    """Utility commands"""
    pass


@utils.command()
@click.option("--port", type=int, default=8000)
@click.option("--debug", is_flag=True, default=False)
@click.option("--repaint", is_flag=True, default=False)
@click.pass_context
def server(ctx: click.Context, port: int, debug: bool, repaint: bool) -> None:
    if debug:
        logger.setLevel(logging.DEBUG)

    config_path = os.path.abspath("combine.yml")
    combine = Combine(
        config_path=config_path,
        env="development",
        variables={"base_url": f"http://127.0.0.1:{port}"},
    )

    Server(combine.output_path, Repaint() if repaint else None, port).serve()


@utils.command()
@click.option("--port", type=int, default=8000)
@click.option("--debug", is_flag=True, default=False)
@click.option("--repaint", is_flag=True, default=False)
@click.pass_context
def watch(ctx: click.Context, port: int, debug: bool, repaint: bool) -> None:
    if debug:
        logger.setLevel(logging.DEBUG)

    config_path = os.path.abspath("combine.yml")
    combine = Combine(
        config_path=config_path,
        env="development",
        variables={"base_url": f"http://127.0.0.1:{port}"},
    )

    Watcher(".", combine=combine, repaint=Repaint() if repaint else None).watch()


@utils.command()
@click.option(
    "--style",
    default="default",
    show_default=True,
    type=click.Choice(list(pygments.styles.get_all_styles())),
)
@click.pass_context
@cls_client.track_command(include_kwargs=["style"])
def highlight_css(ctx: click.Context, style: str) -> None:
    """Outputs the CSS which can be customized for highlighted code"""
    for line in (
        pygments.formatters.HtmlFormatter(style=style).get_style_defs().splitlines()
    ):
        click.echo(f".highlight {line}")


@cli.command()
def update() -> None:
    """Update your version of combine"""
    barrel.update("combine")


if __name__ == "__main__":
    cli()



import os

from .base import Check
from .issues import Issues, Issue


class FaviconCheck(Check):
    def __init__(self, site_dir: str) -> None:
        self.site_dir = site_dir

    def run(self) -> Issues:
        issues = Issues()

        if not os.path.exists(os.path.join(self.site_dir, "favicon.ico")):
            issues.append(
                Issue(
                    type="favicon-missing",
                    description="Your site should have a Favicon at /favicon.ico.",
                )
            )

        return issues



import bs4
from .base import Check
from .issues import Issues, Issue


class ImgAltCheck(Check):
    def __init__(self, html_soup: bs4.BeautifulSoup) -> None:
        self.html_soup = html_soup

    def run(self) -> Issues:
        issues = Issues()

        for img in self.html_soup.findAll("img"):
            alt = img.get("alt")
            if alt is None:
                issues.append(
                    Issue(
                        type="image-alt-missing",
                        description="All <img> tags should have alt text describing the image, or be set to an empty string (`"
                        "`)",
                        context={"element": str(img)},
                    )
                )

        return issues



from typing import Any, Callable
from jinja2 import nodes, pass_context
from jinja2.parser import Parser
from jinja2.ext import Extension
from markupsafe import Markup

import markdown
from markdown.extensions.codehilite import CodeHiliteExtension


class MarkdownExtension(Extension):
    tags = set(["markdown"])

    def parse(self, parser: Parser) -> nodes.Node:
        lineno = next(parser.stream).lineno
        body = parser.parse_statements(("name:endmarkdown",), drop_needle=True)
        return nodes.CallBlock(self.call_method("_support"), [], [], body).set_lineno(
            lineno
        )

    def _support(self, caller: Callable) -> str:
        """Helper callback."""
        markdown_content = caller()
        # jinja will have escaped by default, so we want to unescape
        # for now and leave that to markdown rendering
        markdown_content = Markup(markdown_content).unescape()
        return markdown_to_html(markdown_content)


@pass_context
def markdown_filter(ctx: dict, value: str) -> Markup:
    html_content = markdown_to_html(value)
    return Markup(html_content)


def markdown_to_html(markdown_content: str) -> str:
    html_content = markdown.markdown(
        markdown_content,
        extensions=[
            "markdown.extensions.fenced_code",
            CodeHiliteExtension(css_class="highlight"),
            "markdown.extensions.tables",
            "markdown.extensions.toc",
        ],
    )
    return html_content



import datetime
import os
import time
from http.server import HTTPServer as BaseHTTPServer, SimpleHTTPRequestHandler

import click
from watchdog.observers import Observer
from watchdog.events import (
    FileSystemEventHandler,
    FileCreatedEvent,
    DirModifiedEvent,
    FileDeletedEvent,
    FileMovedEvent,
    DirDeletedEvent,
    DirMovedEvent,
)

from .exceptions import BuildError
from .files.ignored import IgnoredFile
from .logger import logger

from typing import TYPE_CHECKING, Optional, Callable, List, Any, Tuple, Type

if TYPE_CHECKING:
    from .core import Combine
    from repaint import Repaint


class Watcher:
    def __init__(
        self, path: str, combine: "Combine", repaint: Optional["Repaint"] = None
    ) -> None:
        self.path = path
        self.observer = Observer()
        self.event_handler = EventHandler(combine, repaint)

    def watch(self, while_running_func: Callable = None) -> None:
        self.observer.schedule(self.event_handler, self.path, recursive=True)
        self.observer.start()
        try:
            if while_running_func is not None:
                while_running_func()
            else:
                while True:
                    time.sleep(1)
        except (Exception, KeyboardInterrupt) as e:
            print(e)
            self.observer.stop()
        self.observer.join()


class EventHandler(FileSystemEventHandler):
    def __init__(
        self,
        combine: "Combine",
        repaint: Optional["Repaint"],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.combine = combine
        self.repaint = repaint

        # To dedupe modified events
        self.last_valid_event: Optional[FileSystemEventHandler] = None
        self.last_event_time: Optional[datetime.datetime] = None

        super().__init__(*args, **kwargs)

    def should_ignore_path(self, event_path: str) -> bool:
        ignore_dirs = (
            "node_modules",
            ".cache",
            ".venv",
            "env",
            ".git",
        )

        for p in os.path.abspath(event_path).split(os.sep):
            if p in ignore_dirs:
                return True

        ignore_extensions = (".crdownload",)

        if os.path.splitext(event_path)[1] in ignore_extensions:
            return True

        return False

    def is_duplicate_event(self, event: FileSystemEventHandler) -> bool:
        return (
            self.last_valid_event is not None
            and self.last_event_time is not None
            and event == self.last_valid_event
            and (datetime.datetime.now() - self.last_event_time)
            < datetime.timedelta(seconds=0.1)
        )

    def valid_event(self, event: FileSystemEventHandler) -> None:
        """When an event is processed, call this to make sure we can
        deduplicate file events that sometimes trigger twice"""
        self.last_valid_event = event
        self.last_event_time = datetime.datetime.now()

    def on_any_event(self, event: FileSystemEventHandler) -> None:
        logger.debug("Event: %s", event)

        # if a file was moved or something, we only care about the destination
        event_path = event.dest_path if hasattr(event, "dest_path") else event.src_path

        if self.should_ignore_path(event_path):
            logger.debug("Ignoring path: %s", event_path)
            return

        if self.is_duplicate_event(event):
            logger.debug("Duplicate event: %s", event)
            return

        if self.combine.is_in_output_path(event_path):
            _, ext = os.path.splitext(event_path)
            if ext in (".css", ".img", ".js") and self.repaint:
                output_relative_path = os.path.relpath(
                    event_path, self.combine.output_path
                )
                logger.debug("Repainting output path: %s", output_relative_path)
                self.repaint.reload_assets([output_relative_path])
            else:
                logger.debug("Ignoring output path: %s", event_path)
            return

        for step in self.combine.config.steps:
            matched_pattern = step.watch_pattern_match(event_path)
            if matched_pattern:
                click.secho(
                    f"Running step for matching {matched_pattern}",
                    bold=True,
                    color=True,
                )
                result = step.run_process(check=False)
                if result.returncode != 0:
                    click.secho(
                        "There was an error running a user command.",
                        fg="red",
                        color=True,
                    )

        if os.path.abspath(event_path) == os.path.abspath(self.combine.config_path):
            click.secho(
                f"{self.combine.config_path} {event.event_type}: reloading combine and rebuilding site",
                bold=True,
                color=True,
            )
            self.reload_combine()
            self.rebuild_site()
            self.valid_event(event)
            return

        content_relative_path = self.combine.content_relative_path(
            os.path.abspath(event_path)
        )

        if content_relative_path:

            if isinstance(event, (FileCreatedEvent, DirModifiedEvent)):
                # Reload first, so we know about any new files
                self.reload_combine()

            if isinstance(
                event,
                (FileDeletedEvent, DirDeletedEvent, DirMovedEvent, FileMovedEvent),
            ):
                click.secho(
                    f"{content_relative_path} {event.event_type}: ",
                    nl=False,
                    bold=True,
                    color=True,
                )
                click.echo("Rebuilding entire site")
                self.reload_combine()
                self.rebuild_site()
                self.valid_event(event)
                return

            files = self.combine.get_related_files(content_relative_path)

            if files and all([type(f) == IgnoredFile for f in files]):
                return

            click.secho(
                f"{content_relative_path} {event.event_type}: ",
                nl=False,
                bold=True,
                color=True,
            )

            if len(files) == 1:
                click.echo(f"Rebuilding {files[0].content_relative_path}")
            elif not files:
                click.echo("Rebuliding entire site")
            else:
                click.echo(f"Rebuilding {len(files)} files")
            self.rebuild_site(only_paths=[x.path for x in files])
            self.valid_event(event)

    def reload_combine(self) -> None:
        try:
            self.combine.reload()
            if self.repaint:
                self.repaint.reload()
        except Exception as e:
            logger.error("Error reloading", exc_info=e)
            click.secho("There was an error! See output above.", fg="red", color=True)

    def rebuild_site(self, only_paths: List[str] = []) -> None:
        try:
            self.combine.build(only_paths)
            if self.repaint:
                self.repaint.reload()
        except BuildError:
            click.secho("Build error (see above)", fg="red", color=True)
        except Exception as e:
            logger.error("Error building", exc_info=e)
            click.secho("There was an error! See output above.", fg="red", color=True)


class Server:
    def __init__(
        self, path: str, repaint: Optional["Repaint"] = None, port: int = 8000
    ) -> None:
        self.path = path
        self.port = port
        self.repaint = repaint
        self.httpd = HTTPServer(self.path, ("", self.port), repaint)

    def serve(self) -> None:
        self.httpd.serve_forever()


class HTTPHandler(SimpleHTTPRequestHandler):
    """This handler uses server.base_path instead of always using os.getcwd()"""

    @property
    def repaint(self) -> Optional["Repaint"]:
        return self.server.repaint  # type: ignore

    @property
    def base_path(self) -> str:
        return self.server.base_path  # type: ignore

    def inject_repaint(self) -> None:
        if not self.repaint:
            return

        path = self.translate_path(self.path)
        path_extension = os.path.splitext(path)[1]

        if path_extension not in ("", ".html"):
            return

        if path_extension == "":
            # Guess that there is an index.html
            path = os.path.join(path, "index.html")

        if os.path.exists(path):
            with open(path, "r") as f:
                contents = f.read()

            repaint_script = self.repaint.script_tag
            if repaint_script not in contents:
                contents = contents.replace("</body>", repaint_script + "</body>")

            with open(path, "w") as f:
                f.write(contents)

    def do_GET(self) -> None:
        self.inject_repaint()
        return super().do_GET()

    def translate_path(self, path: str) -> str:
        path = super().translate_path(path)
        relpath = os.path.relpath(path, os.getcwd())
        fullpath = os.path.join(self.base_path, relpath)
        return fullpath

    def log_message(self, format: str, *args: Any) -> None:
        """Disable logging"""
        return


class HTTPServer(BaseHTTPServer):
    """The main server, you pass in base_path which is the path you want to serve requests from"""

    def __init__(
        self,
        base_path: str,
        server_address: Tuple[str, int],
        repaint: Optional["Repaint"],
        RequestHandlerClass: Type[HTTPHandler] = HTTPHandler,
    ) -> None:
        self.base_path = base_path
        self.repaint = repaint
        super().__init__(server_address, RequestHandlerClass)



import bs4

from .base import Check
from .issues import Issues, Issue


class MixedContentCheck(Check):
    def __init__(self, html_soup: bs4.BeautifulSoup) -> None:
        self.html_soup = html_soup

    def run(self) -> Issues:
        issues = Issues()

        to_check = {
            "img": {"attr": "src"},
            "link": {"attr": "href", "ignore": {"rel": "profile"}},
            "iframe": {"attr": "src"},
            # TODO missing script? but src needs to be optional (could be inline)
        }

        for tag, cfg in to_check.items():
            for el in self.html_soup.findAll(tag):
                if any(
                    [
                        el.get(k, [None])[0] == v
                        for k, v in cfg.get("ignore", {}).items()  # type: ignore
                    ]
                ):
                    continue

                attr = cfg["attr"]  # type: ignore

                value = el.get(attr)

                if value.startswith("http:"):
                    issues.append(
                        Issue(
                            type="https-mixed-content",
                            description="Any linked resources (CSS, img, iframes) should be linked via HTTPS.",
                            context={"element": str(el)},
                        )
                    )

        return issues



import os
import traceback

import jinja2
from .html import HTMLFile
from .utils import create_parent_directory

from typing import Any


class ErrorFile(HTMLFile):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.error = kwargs.pop("error")
        super().__init__(*args, **kwargs)

    def _render_to_output(
        self, output_path: str, jinja_environment: jinja2.Environment
    ) -> str:
        target_path = os.path.join(output_path, self.output_relative_path)
        create_parent_directory(target_path)

        template = jinja_environment.get_template("error.template.html")

        with open(target_path, "w+") as f:
            if hasattr(self.error, "source"):
                context_lines = "\n".join(
                    self.error.source.splitlines()[
                        max(self.error.lineno - 3, 0) : self.error.lineno + 3
                    ]
                )
            else:
                context_lines = ""

            f.write(
                template.render(
                    error=self.error,
                    relative_path=self.content_relative_path,
                    context_lines=context_lines,
                    excinfo=traceback.format_exc(),
                )
            )

        return target_path



from urllib.parse import urlparse
from jinja2 import pass_context
from .exceptions import MissingVariableError


@pass_context
def absolute_url(ctx: dict, value: str) -> str:
    try:
        base_url = ctx["base_url"]
    except KeyError:
        raise MissingVariableError("base_url")

    # TODO any way to use jinja error and point to line number etc?

    if not base_url:
        raise Exception("absolute_url error: base_url can't be empty")

    if not value:
        raise Exception("absolute_url error: url argument can't be empty")

    if base_url.endswith("/") and value.startswith("/"):
        url = base_url + value.lstrip("/")

    if not base_url.endswith("/") and not value.startswith("/"):
        url = base_url + "/" + value

    url = base_url + value

    parsed = urlparse(url)
    if not parsed.scheme or not parsed.netloc:
        raise Exception(f"absolute_url error: absolute url doesn't look valid\n{url}")

    return url



import os
from shutil import copyfile
from typing import List, TYPE_CHECKING

import jinja2
from ..checks.issues import Issues
from ..checks.file_size import FileSizeCheck
from .utils import create_parent_directory
from ..checks.base import Check

if TYPE_CHECKING:
    from combine.core import ContentDirectory


class File:
    def __init__(self, path: str, content_directory: "ContentDirectory") -> None:
        self.path = path
        self.references: List[str] = []
        self.content_directory = content_directory
        self.content_relative_path = os.path.relpath(
            self.path, self.content_directory.path
        )

        self.root, self.extension = os.path.splitext(self.content_relative_path)
        self.root_parts = os.path.split(self.root)
        self.name_without_extension = self.root_parts[-1]

        self.output_relative_path = self._get_output_relative_path()

    def _get_output_relative_path(self) -> str:
        return self.content_relative_path

    def load(self, jinja_environment: jinja2.Environment) -> None:
        """Load properties that can vary depending on content of the file"""
        self.references = []

    def render(self, output_path: str, jinja_environment: jinja2.Environment) -> None:
        self.output_path = self._render_to_output(output_path, jinja_environment)

    def _render_to_output(
        self, output_path: str, jinja_environment: jinja2.Environment
    ) -> str:
        target_path = os.path.join(output_path, self.output_relative_path)
        create_parent_directory(target_path)

        if os.path.exists(target_path):
            os.remove(target_path)

        copyfile(self.path, target_path)

        return target_path

    def check_output(self) -> Issues:
        issues = Issues()

        for check in self.get_checks():
            for issue in check.run():
                issues.append(issue)

        if issues:
            issues.print(f"Issues in {self.content_relative_path}")

        return issues

    def get_checks(self) -> List[Check]:
        if self.output_path:
            # Not all files have an output
            return [
                FileSizeCheck(path=self.output_path),
            ]

        return []



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['combine', 'combine.checks', 'combine.files', 'combine.jinja']

package_data = \
{'': ['*'], 'combine': ['base_content/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'Pygments>=2.6.1,<3.0.0',
 'barrel>=0.3.0,<0.4.0',
 'beautifulsoup4>=4.9.1,<5.0.0',
 'click>7.1.2',
 'cls-client>=1.2.0,<2.0.0',
 'honcho>=1.1.0,<2.0.0',
 'markdown>=3.2.2,<4.0.0',
 'python-frontmatter>=1.0.0,<2.0.0',
 'repaint>=0.1.0,<0.2.0',
 'typing-extensions>=4.3.0,<5.0.0',
 'watchdog>=2.1.6,<3.0.0']

entry_points = \
{'console_scripts': ['combine = combine.cli:cli']}

setup_kwargs = {
    'name': 'combine',
    'version': '2.12.0',
    'description': 'A straightforward static site builder.',
    'long_description': '# Combine\n\n**Build a straightforward marketing or documentation website with the power of [Jinja](http://jinja.pocoo.org/).\nNo fancy JavaScript here &mdash; this is just like the good old days.**\n\nPut your site into the "content" directory and Combine will:\n\n- Render files using Jinja\n- Create pretty URLs ("file-system routing")\n- Inject variables\n- Run custom build steps (like building Tailwind)\n\nMost sites follow a simple pattern.\n\nCreate a `base.template.html`:\n\n```html+jinja\n<!DOCTYPE html>\n<html lang="en">\n<head>\n    <meta charset="UTF-8">\n    <meta name="viewport" content="width=device-width, initial-scale=1.0">\n    <title>My site</title>\n</head>\n<body>\n    {% block content %}{% endblock %}\n</body>\n</html>\n```\n\nAnd use it (ex. `pricing.html`):\n\n```html+jinja\n{% extends "base.template.html" %}\n\n{% block content %}\n<div class="pricing">\n    <div class="flex">\n        ...\n    </div>\n</div>\n{% endblock %}\n```\n\nIn the end, you get a static HTML site that can be deployed almost anywhere.\n',
    'author': 'Dropseed',
    'author_email': 'python@dropseed.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://combine.dropseed.dev',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



import jinja2
from ..jinja.references import get_references_in_path
from .ignored import IgnoredFile


class TemplateFile(IgnoredFile):
    def load(self, jinja_environment: jinja2.Environment) -> None:
        self.references = get_references_in_path(self.path, jinja_environment)

    def _render_to_output(
        self, output_path: str, jinja_environment: jinja2.Environment
    ) -> str:
        return ""



import bs4
import os
from urllib.parse import urljoin

from .base import Check
from .issues import Issues, Issue


class InternalLinkBrokenCheck(Check):
    def __init__(
        self, html_soup: bs4.BeautifulSoup, file_path: str, output_dir: str
    ) -> None:
        self.html_soup = html_soup
        self.file_path = file_path
        self.output_dir = output_dir

    def run(self) -> Issues:
        issues = Issues()

        types = {
            "img": "src",
            "script": "src",
            "a": "href",
            "link": "href",
        }

        skip_prefixes = (
            "//",
            "http:",
            "https:",
            "tel:",
            "sms:",
            "mailto:",
            "ftp:",
            "file:",
            "#",
        )

        for nodeType, nodeAttribute in types.items():
            for node in self.html_soup.findAll(nodeType):
                value = node.get(nodeAttribute)

                if value:
                    # Remove whitespace on both ends
                    value = value.strip()

                if value and "?" in value:
                    # Remove query params too (style.css?v=1.0)
                    value = value.split("?")[0]

                if not value:
                    # Skip empty ones for now, not our responsibility
                    continue

                should_skip = False

                for p in skip_prefixes:
                    if value.startswith(p):
                        should_skip = True
                        break

                if should_skip:
                    continue

                if value.startswith("/"):
                    # remove the leading / and join to output_dir
                    output_path = urljoin(self.output_dir, value[1:])
                else:
                    output_path = urljoin(self.file_path, value)

                _, ext = os.path.splitext(output_path)
                if not ext:
                    # If it's a directory, pretend we're a webserver and
                    # look for index.html
                    if not output_path.endswith("/"):
                        output_path += "/"
                    output_path = urljoin(output_path, "index.html")

                # TODO if not in output_dir, that's an error ("../../../ that takes you out of combine")

                if not os.path.exists(output_path):
                    issues.append(
                        Issue(
                            type="internal-link-broken",
                            description="You have a link that doesn't point to an existing file.",
                            context={
                                "element": str(node),
                                "target_path": os.path.relpath(output_path),
                            },
                        )
                    )

        return issues
